'''Spiral: SPlitters for IdentifieRs: A Library

Natural language processing (NLP) methods are increasingly being applied to
source code analysis for various purposes.  The methods rely on terms
(identifiers and other textual strings) extracted from program source code
and comments.  The methods often work better if, instead of raw identifiers,
real words are used as features; that is, "get" and "integer" are often better
features for NLP tools than the string "getInteger".  This leads to the
need for automated methods for splitting identifiers of classes, functions,
variables, and other entities into word-like constituents.  A number of
methods have been proposed an implemented to perform identifier splitting.

Spiral is a Python 3 package that implements numerous identifier splitting
algorithms.  It provides some basic naive splitting algorithms, such as a
straightforward camel-case splitter, as well as more elaborate heuristic
splitters, such as an algorithm we call Ronin (originally based on an
algorithm called Samurai by Enslen, Hill, Pollock and Vija-Shanker [2009]).

Authors
-------

Michael Hucka <mhucka@caltech.edu>

Copyright
---------

Copyright (c) 2017 by the California Institute of Technology.  This software
was developed as part of the CASICS project, the Comprehensive and Automated
Software Inventory Creation System.  For more, visit http://casics.org.
'''

from .__version__ import __version__, __title__, __description__, __url__
from .__version__ import __author__, __email__
from .__version__ import __license__, __copyright__

# Supporting modules.
from .frequencies import frequencies_from_csv_file, frequencies_from_pickle
from .frequencies import save_frequencies_to_pickle

# Simple splitters.
from .simple_splitters import delimiter_split, digit_split, pure_camelcase_split
from .simple_splitters import safe_simple_split, simple_split, elementary_split
from .simple_splitters import heuristic_split

# Advanced splitters.
from . import samurai
from . import ronin



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import re
import textdistance
import itertools
import functools
import numpy as np

from .lemmatizer import Lemmatizer

DLDis = textdistance.DamerauLevenshtein()
LCSSeq = textdistance.LCSSeq()

class PairChecker:
    
    @staticmethod
    @functools.lru_cache(maxsize=10000)
    def normalize(term):
        lemma = Lemmatizer.lemmatize_noun(term)
        name = lemma.replace("-", " ")
        name = re.sub(r"\s+", " ", name)
        return name

    @staticmethod
    def check_synonym(long_term, short_term, lemma=True, max_dis=2, threshold=0.25):
        if long_term.isupper() and short_term.isupper():
            return False
        # if long_term.lower().rstrip("s") == short_term.lower().rstrip("s"):
        #     return True
        if lemma:
            long_term = PairChecker.normalize(long_term)
            short_term = PairChecker.normalize(short_term)

        if long_term == short_term:
            return True
        if long_term.replace(" ", "") == short_term.replace(" ", ""):
            return True

        long_words = long_term.split()
        short_words = short_term.split()

        if len(long_words) != len(short_words):
            return False
        for word1, word2 in zip(long_words, short_words):
            if word1[0] != word2[0]:
                return False
            if re.findall(r'[0-9]+', word1) != re.findall(r'[0-9]+', word2):
                return False
            if DLDis.distance(word1, word2) > max_dis or DLDis.normalized_distance(word1, word2) >= threshold:
                return False
        return True

    @staticmethod
    def __check_word_word(long_word, short_word, thres=2/3):
        if re.match(r'^[0-9]+$', long_word) or re.match(r'^[0-9]+$', long_word):
            return False
        if long_word[0] != short_word[0]:
            return False
        if len(short_word) / len(long_word) >= thres:
            return False
        # logging.info(long_word, short_word, LCSSeq.similarity(long_word, short_word))
        if LCSSeq.similarity(long_word, short_word) != len(short_word):
            return False
        
        # check prefix
        if long_word.startswith(short_word):
            return True
        # check acronym
        if len(set(short_word) & {"a", "e", "i", "o", "u"}) == 0:
            return True
        return False

    @staticmethod
    def __check_phrase_word(phrase, word):
        if phrase[0] != word[0]:
            return False
        words = phrase.split()
        if len(word) < len(words):
            return False
        if len(words) == 1:
            return PairChecker.__check_word_word(phrase, word)
        else:
            if words[0] == word or len(words) != len(word):
                return False
            for indices in itertools.combinations([i for i in range(1, len(word), 1)], len(words) - 1):
                # logging.info(indices)
                copied_words = words.copy()
                indices = (0, ) + indices + (len(word), )
                # logging.info(indices)
                pairs = [(beg, end) for (beg, end) in zip(indices[:-1], indices[1:])]
                beg, end = pairs.pop(0)
                cur_chars = word[beg:end]
                
                cur_word = copied_words.pop(0)
                # logging.info(cur_chars, cur_word)
                if cur_chars[0] == cur_word[0]:
                    if not PairChecker.__check_word_word(cur_word, cur_chars):
                        continue
                elif cur_word[0] in set("aeiou") and len(cur_word) > 1 and cur_chars[0] == cur_word[1]:
                    if end != 1:
                        continue
                else:
                    continue

                for (beg, end), cur_word in zip(pairs, copied_words):
                    cur_chars = word[beg:end]
                    # logging.info(cur_chars, cur_word)
                    if not PairChecker.__check_word_word(cur_word, cur_chars):
                        break
                else:
                    return True

            return False

    @staticmethod
    def check_abbr(long_term, short_term, lemma=True):
        if lemma:
            long_term = PairChecker.normalize(long_term)
            short_term = PairChecker.normalize(short_term)

        # logging.info(long_term, short_term)

        if len(short_term.split()) == 1:
            return PairChecker.__check_phrase_word(long_term, short_term)
        else:
            short_words = short_term.split()
            long_words = long_term.split()
            if len(long_words) < len(short_words):
                return False
            while len(short_words) > 0 and short_words[0] == long_words[0]:
                short_words.pop(0)
                long_words.pop(0)
            while len(short_words) > 0 and short_words[-1] == long_words[-1]:
                short_words.pop(-1)
                long_words.pop(-1)
            if len(short_words) == 1:
                return PairChecker.__check_phrase_word(" ".join(long_words), short_words[0])
            elif len(short_words) == len(long_words):
                for word1, word2 in zip(long_words, short_words):
                    if not PairChecker.__check_word_word(word1, word2):
                        return False
                return True
            return False


'''
samurai: implementation of the Samurai algorithm for identifier splitting

Introduction
------------

Natural language processing (NLP) methods are increasingly being applied to
source code analysis for various purposes.  The methods rely on terms
(identifiers and other textual strings) extracted from program source code
and comments.  The methods often work better if, instead of raw identifiers,
real words are used as features; that is, "get" and "integer" are often better
features for NLP tools than the string "getInteger".  This leads to the
need for automated methods for splitting identifiers of classes, functions,
variables, and other entities into word-like constituents.  A number of
methods have been proposed an implemented to perform identifier splitting.

Samurai is a Python module that implements a splitter known as Samurai,
published by Eric Enslen, Emily Hill, Lori Pollock and K. Vijay-Shanker from
the University of Delaware in 2009.

  Enslen, E., Hill, E., Pollock, L., & Vijay-Shanker, K. (2009).  Mining
  source code to automatically split identifiers for software analysis.  In
  Proceedings of the 6th IEEE International Working Conference on Mining
  Software Repositories (MSR'09) (pp. 71-80).

The implementation in the present Python module was created by the author
from the description of the algorithm published in the paper.  Assuming my
interpretation of the Enslen et al.'s algorithm is correct, this
implementation is faithful to the algorithm described in the paper.

Usage
-----

Note: for fastest performance, using the optimization options provided by the
Python interpreter (i.e., the options -O or -OO).

The simplest way to use this module in a Python program is to import the
module from the Spiral package and then call samurai.split() with a
string. For example,

    from spiral import samurai
    result = samurai.split('someidentifier')

The function split(...) produces a list of strings as its output.  It will do
its best-guess at how a given identifier should be split using a built-in
table of term frequencies, prefixes, and suffixes.  The built-in table of
term frequencies came from an analysis of 20,000 randomly selected software
projects in GitHub that contained at least one Python source code file.

Samurai performs best when it is provided with project-specific tables of
term frequencies.  If you have a set of frequencies for your project, you can
supply them to the function init(...) once, before making any calls to
split().  (This is done so that the cost of computations involved in the
frequencies is only done once.)  Use the keyword argument 'local_freq'

    from spiral import samurai
    samurai.init(local_freq=myfrequencies)
    result = samurai.split('someidentifier')

The format of the value given to 'local_freq' should be a dictionary where
the keys are the terms and the values are integers representing the number of
times the term appears in the code base.  A term that appears once will have
a value of 1, a term that appears twice will have a value of 2, etc.  Some
terms for large code bases can have values that go into the hundreds of
thousands; this is okay, as Samurai will normalize the values.

You can also change the global frequency table that Samurai uses.  This can
be passed using the keyword argument 'global_freq' to init(...):

    samurai.init(global_freq=globalfrequencies, local_freq=projectfrequencies)

Note that generating the global table is not a trivial undertaking.  It
requires mining source code repositories to extract identifiers of classes,
functions, variables, etc., splitting them to produce individual identifier
fragments, and finally curating the result manually.

Tracing the algorithm
---------------------

To print what the splitter is doing while processing a given string, you can
turn on logging by setting the logging level to "logging.DEBUG" before
importing the samurai module.  Here is an example:

  import logging
  logging.basicConfig(level=logging.DEBUG, format='')
  from spiral import samurai

Logging will be printed to the standard output stream.  Note: this will only
work if you do NOT use the -O option to the Python interpreter.

Optimization
------------

This module wraps all logging code with the Python __debug__ compile-time
symbol, so that you can use the -O flag to Python to make it omit the code
completely.  This is more efficient than simply setting the logging level.

Authors
-------

Michael Hucka <mhucka@caltech.edu>

Copyright
---------

Copyright (c) 2017 by the California Institute of Technology.  This software
was developed as part of the CASICS project, the Comprehensive and Automated
Software Inventory Creation System.  For more, visit http://casics.org.
'''

import math
import os
import re
import sys

try:
    sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
except:
    sys.path.append("..")

try:
    from .frequencies import *
    from .simple_splitters import simple_split
    from .constants import prefixes, suffixes
except:
    from frequencies import *
    from simple_splitters import simple_split
    from constants import prefixes, suffixes


# Global constants.
# .............................................................................

# If the caller doesn't supply a frequency table, we use a default one.
# The file needs to be kept in the same directory as this Python module file.

try:
    _DEFAULT_FREQ_DIR = os.path.dirname(__file__)
except:
    _DEFAULT_FREQ_DIR = os.path.dirname('.')
_DEFAULT_FREQ_PICKLE = os.path.join(_DEFAULT_FREQ_DIR, 'data/frequencies.pklz')


# Main class
# .............................................................................

class Samurai(object):
    _global_freq = None
    _local_freq  = None

    def init(self, local_freq=None, global_freq=None):
        '''Initialize internal frequency files for the Samurai split() function.
        Parameter 'local_freq' should be a dictionary where the keys are the
        terms and the values are integers representing the number of times
        the term appears in the code base.  A term that appears once will
        have a value of 1, a term that appears twice will have a value of 2,
        etc.  Similarly, 'glocal_freq' should be a dictionary for global
        frequencies of all terms in a large sample of code bases.
        '''
        if local_freq:
            self._local_freq = local_freq
        if global_freq:
            self._global_freq = global_freq
        if not self._global_freq:
            if os.path.exists(_DEFAULT_FREQ_PICKLE):
                self._global_freq = frequencies_from_pickle(_DEFAULT_FREQ_PICKLE)
            else:
                raise ValueError('Cannot read default frequencies pickle file "{}"'
                                 .format(_DEFAULT_FREQ_PICKLE))


    def split(self, identifier):
        '''Split the given identifier string at token boundaries using a
        combination of heuristics and token frequency tables.  This function
        produces a list of strings as its output.  The Samurai algorithm is
        described by Eric Enslen, Emily Hill, Lori Pollock and
        K. Vijay-Shanker in a 2009 paper titled "Mining source code to
        automatically split identifiers for software analysis" (Proceedings
        of the 6th IEEE International Working Conference on Mining Software
        Repositories, 2009).

        If you have a set of frequencies for your project, you can supply
        them to the function init(...) once, before making calls to split().
        This function will otherwise use a built-in global table of
        frequencies.
        '''
        splits = []
        if not self._global_freq:
            self.init()
        score = self._generate_scoring_function()
        for s in simple_split(identifier):
            # Look for upper-to-lower case transitions
            transition = re.search(r'[A-Z][a-z]', s)
            if not transition:
                parts = [s]
            else:
                i = transition.start(0)
                if i > 0:
                    camel_score = score(s[i:])
                else:
                    camel_score = score(s)
                alt = s[i+1:]
                alt_score = score(alt)
                if camel_score > math.sqrt(alt_score):
                    if i > 0:
                        parts = [s[0:i], s[i:]]
                    else:
                        parts = [s]
                else:
                    parts = [s[0:i+1], s[i+1:]]
            splits = splits + parts

        results = []
        for token in splits:
            results = results + self._same_case_split(token, score, score(token))
        return results


    def _same_case_split(self, s, score, score_ns=0.0000005):
        if len(s) < 2:
            return [s]

        split     = None
        n         = len(s)
        i         = 0
        max_score = -1
        threshold = max(score(s), score_ns)
        while i < n:
            left       = s[0:i]
            right      = s[i:n]
            score_l    = score(left)
            score_r    = score(right)
            prefix     = self._is_prefix(left) or self._is_suffix(right)
            to_split_l = math.sqrt(score_l) > threshold
            to_split_r = math.sqrt(score_r) > threshold

            if not prefix and to_split_l and to_split_r:
                if (score_l + score_r) > max_score:
                    max_score = score_l + score_r
                    split = [left, right]
            elif not prefix and to_split_l:
                tmp = self._same_case_split(right, score, score_ns)
                if tmp[0] != right:
                    split = [left] + tmp
            i += 1
        result = split if split else [s]
        return result


    def _generate_scoring_function(self):
        '''Returns a closure that computes a score using the frequency tables
        with which the splitter has been initialized.
        '''
        if self._local_freq:
            log_all_freq = math.log10(sum(self._local_freq.values()))
            def scoring_function(w):
                local_f  = self._local_freq[w] if w in self._local_freq else 0
                global_f = self._global_freq[w] if w in self._global_freq else 0
                if w not in self._global_freq:
                    return local_f
                if w not in self._local_freq:
                    return global_f/log_all_freq
                else:
                    return local_f + global_f/log_all_freq
        else:
            def scoring_function(w):
                return self._global_freq[w] if w in self._global_freq else 0
        return scoring_function


    def _is_prefix(self, s):
        return s.lower() in prefixes


    def _is_suffix(self, s):
        return s.lower() in suffixes


# Module entry points.
# .............................................................................

_SAMURAI_INSTANCE = Samurai()

# The following are the functions we tell people to use:

init  = _SAMURAI_INSTANCE.init
split = _SAMURAI_INSTANCE.split



#!/usr/bin/env python
# -*- encoding: utf-8 -*-


from pathlib import Path

DEFAULT_DICT_DIR = Path(__file__).parent / "dicts"
DEFAULT_DICT = {
}

DEFAULT_METHOD_CORPUS = Path(__file__).parent / "corpus/normal.methods"
DEFAULT_FIELD_CORPUS = Path(__file__).parent / "corpus/normal.fields"



def load_default_dict():
    with (DEFAULT_DICT_DIR /"dictionary-allwords").open("r", encoding="utf-8") as f:
        vocab = {line.strip() for line in f if line.strip() != ""}
    DEFAULT_DICT["VOCAB"] = vocab
    with (DEFAULT_DICT_DIR /"preposition").open("r", encoding="utf-8") as f:
        prep = {line.strip() for line in f if line.strip() != ""}
    DEFAULT_DICT["PREP"] = prep
    with (DEFAULT_DICT_DIR /"quant").open("r", encoding="utf-8") as f:
        quant = {line.strip() for line in f if line.strip() != ""}
    DEFAULT_DICT["QUAN"] = quant
    with (DEFAULT_DICT_DIR /"irregV").open("r", encoding="utf-8") as f:
        irregV = {line.strip() for line in f if line.strip() != ""}
    DEFAULT_DICT["IRRV"] = irregV
    with (DEFAULT_DICT_DIR /"adjective").open("r", encoding="utf-8") as f:
        adj = {line.strip() for line in f if line.strip() != ""}
    DEFAULT_DICT["ADJ"] = adj
    with (DEFAULT_DICT_DIR /"adverb").open("r", encoding="utf-8") as f:
        adv = {line.strip() for line in f if line.strip() != ""}
    DEFAULT_DICT["ADV"] = adv
    with (DEFAULT_DICT_DIR /"noun").open("r", encoding="utf-8") as f:
        noun = {line.strip() for line in f if line.strip() != ""}
    DEFAULT_DICT["NOUN"] = noun
    with (DEFAULT_DICT_DIR /"verb").open("r", encoding="utf-8") as f:
        verb = {line.strip() for line in f if line.strip() != ""}
    DEFAULT_DICT["VERB"] = verb
    with (DEFAULT_DICT_DIR /"pronoun").open("r", encoding="utf-8") as f:
        pronoun = {line.strip() for line in f if line.strip() != ""}
    DEFAULT_DICT["PRON"] = pronoun
    with (DEFAULT_DICT_DIR /"n-abbr").open("r", encoding="utf-8") as f:
        nabbr = {line.strip() for line in f if line.strip() != ""}
    DEFAULT_DICT["NABBR"] = nabbr
    with (DEFAULT_DICT_DIR /"participle").open("r", encoding="utf-8") as f:
        pp = {line.strip() for line in f if line.strip() != ""}
    DEFAULT_DICT["PP"] = pp


load_default_dict()





#coding = utf-8

from pathlib import Path
import json

import jpype

DEFAULT_JAR = str(Path(__file__).parent / "CodeAnalysis.jar")

class DependencyAnalyzer:
    def __init__(self, jar_path=DEFAULT_JAR):
        jpype.startJVM(jpype.getDefaultJVMPath(), "-ea", f"-Djava.class.path={jar_path}")
        self.CFGBuilder = jpype.JClass('codeintelligence.codeanalysis.extractors.dependence.cfg.CFGBuilder')
        self.PDGBuilder = jpype.JClass('codeintelligence.codeanalysis.extractors.dependence.pdg.PDGBuilder')

    def __del__(self):
        if jpype.isJVMStarted():
            jpype.shutdownJVM()

    def build_cfg(self, code):
        cfgs = self.CFGBuilder.buildWithCode(code)
        return [json.loads(str(cfg.exportJSON())) for cfg in cfgs]

    def build_pdg(self, code):
        pdgs = self.PDGBuilder.buildWithCode(code)
        return [json.loads(str(pdg.exportJSON())) for pdg in pdgs]


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Author  :   Chong Wang 
@Contact :   chongwang18@fudan.edu.cn
@Time    :   2019/12/17
'''

from pathlib import Path
import functools

from .lemmatizer import Lemmatizer
from .dictionary import PREPOSITIONS, PHRASAL_VERBS


class CodeChunker:
    __INSTANCE = None

    def __init__(self):
        pass

    def __call__(self, name, *verbs):
        return self.chunk(name, *verbs)
    
    def complete_verb(self, lemma, words, verb):
        verb_eles = [verb]
        word_num = len(words)
        if lemma in PHRASAL_VERBS and word_num > 0:
            len2verb = {}
            for pred in PHRASAL_VERBS[lemma]:
                length = len(pred.split())
                if length > word_num:
                    continue
                if length not in len2verb:
                    len2verb[length] = set()
                len2verb[length].add(pred)
            for length, preds in len2verb.items():
                candidate = " ".join([lemma] + words[:length - 1])
                if candidate in preds:
                    verb_eles.extend(words[:length - 1])
                    break
        return " ".join(verb_eles)
        
    
    @functools.lru_cache(maxsize=10000)
    def chunk(self, name, *verbs):
        verbs = {Lemmatizer.lemmatize_verb(verb) for verb in verbs}
        words = name.lower().split()
        size = len(words)
        if size == 0:
            return None
        lemmas = [Lemmatizer.lemmatize_verb(word) for word in words]
        
        segments = [[]]
        idx = 0
        while idx < size:
            word = words[idx]
            lemma = lemmas[idx]
            if word in {"and", "or"} and 0 < idx < size - 1 and (lemmas[idx - 1] in verbs or lemmas[idx + 1] in verbs):
                segments.append([(words[idx + 1], lemmas[idx + 1])])
                idx += 2
            else:
                segments[0].append((word, lemma))
                idx += 1
        # logging.info(segments)
        verb2chunks = {}
        for segment in segments:
            verb = None
            chunks = []
            chunk = []

            while len(segment) > 0:
                word, lemma = segment.pop(0)
                if lemma in verbs:
                    verb = self.complete_verb(lemma, [w for w, _ in segment], word)

                    for _ in range(len(verb.split()) - 1):
                        segment.pop(0)
                    if len(chunk) > 0:
                        chunks.append(" ".join(chunk))
                        chunk = []
                    chunks.append(verb)
                elif word in PREPOSITIONS:
                    if len(chunk) > 0:
                        chunks.append(" ".join(chunk))
                        chunk = []
                    chunks.append(word)
                else:
                    chunk.append(word)
            if len(chunk) > 0:
                chunks.append(" ".join(chunk))
            verb2chunks[verb] = chunks

        rests = []
        for verb, chunks in verb2chunks.items():
            for idx, chunk in enumerate(chunks):
                if chunk == verb:
                    rests.append(chunks[idx + 1:])
                    break
            else:
                rests.append([])
        for idx, (verb, chunks) in enumerate(verb2chunks.items()):
            if len(rests[idx]) == 0:
                if idx > 0 and len(rests[idx - 1]) > 0:
                    verb2chunks[verb] += rests[idx - 1]
                elif idx < len(rests) - 1 and len(rests[idx + 1]) > 0:
                    verb2chunks[verb] += rests[idx + 1]

        return verb2chunks

    @classmethod
    def get_inst(cls, *args):
        if cls.__INSTANCE is None:
            cls.__INSTANCE = cls()
        return cls.__INSTANCE

def chunk(name, *verbs):
    return CodeChunker.get_inst().chunk(name, *verbs)



                    



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import re
import functools

import spacy

from .posse.tagger import Tagger

class CodePOS:
    __INSTANCE = None

    def __init__(self):
        self.tagger = Tagger()
        self.init_secondary_tagger()

    def init_secondary_tagger(self):
        nlp = spacy.load("en_core_web_sm", disable=["parser", "ner"])
        nlp.tokenizer.token_match = re.compile(r"[A-Za-z\d]+-[A-Za-z\d]+|'[a-z]+|''|id|Id|ID").match
        self.secondary_tagger = nlp

    def __call__(self, method_name):
        return self.tag(method_name)

    @functools.lru_cache(maxsize=10000)
    def tag(self, method_name):
        name_pos = None
        name_words = method_name.lower().split()

        if len(name_words) == 0:
            return []

        if name_words[0] == "to" and len(name_words) > 1:
            return [("to", "closedlist")] + [(word, "noun") for word in name_words[1:]]

        if name_words[0] == "new" and len(name_words) > 1:
            return [("new", "verb")] + [(word, "noun") for word in name_words[1:]]

        indices = []
        for index, word in enumerate(name_words):
            if word in {"and", "or"}:
                indices.append(index)
        starts = [0] + [idx + 1 for idx in indices]
        ends = indices + [len(name_words)]
        segments = [" ".join(name_words[start:end]).strip() for start, end in list(zip(starts, ends))]
        pos = []
        queue = []
        while len(segments) > 0:
            segment = segments.pop(0)
            if len(segment) == 0:
                continue
            queue.append(segment)
            segment_pos = self.tagger.tag(" ".join(queue), False)
            # logging.info(segment_pos)
            if segment_pos is None:
                break
            word, tag = segment_pos.eles[0]
            if tag == "verb":
                if len(indices) > 0 and len(pos) == indices[0]:
                    pos.append((name_words[indices[0]], "conj"))
            pos.extend(segment_pos.eles)
            queue = []
        if len(indices) > 0 and len(pos) == indices[0]:
            pos.append((name_words[indices[0]], "conj"))
        if len(pos) == len(name_words):
            name_pos = pos

        # tagging_result = self.tagger.tag(method_name, False)
        # if tagging_result is None:
        #     name_words = method_name.split()
        #     indices = []
        #     for index, word in enumerate(name_words):
        #         if word.lower() in {"and", "or"}:
        #             indices.append(index)
        #     starts = [0] + [idx + 1 for idx in indices]
        #     ends = indices + [len(name_words)]
        #     segments = [" ".join(name_words[start:end]).strip() for start, end in list(zip(starts, ends))]
        #     pos = []
        #     queue = []
        #     while len(segments) > 0:
        #         segment = segments.pop(0)
        #         if len(segment) == 0:
        #             continue
        #         queue.append(segment)
        #         segment_pos = self.tagger.tag(" ".join(queue), False)
        #         if segment_pos is None:
        #             break
        #         word, tag = segment_pos.eles[0]
        #         if tag == "verb":
        #             if len(indices) > 0 and len(pos) == indices[0]:
        #                 pos.append((name_words[indices[0]], "conj"))
        #             pos.extend(segment_pos.eles)
        #             queue = []
        #     if len(indices) > 0 and len(pos) == indices[0]:
        #         pos.append((name_words[indices[0]], "conj"))
        #     if len(pos) == len(name_words):
        #         name_pos = pos
        # else:
        #     name_pos = tagging_result.eles
        # print(name_pos)
        if name_pos is None:
            name_pos = []
            for token in self.secondary_tagger(method_name):
                word = token.text
                tag = token.pos_.lower()
                if tag == "adp":
                    tag = "closedlist"
                if tag == "aux":
                    tag = "verb"
                name_pos.append((word, tag))
        return name_pos

    @classmethod
    def get_inst(cls, *args):
        if cls.__INSTANCE is None:
            cls.__INSTANCE = cls()
        return cls.__INSTANCE

def pos(method_name):
    return CodePOS.get_inst().tag(method_name)



# =============================================================================
# @file    __version__.py
# @brief   CASICS Spiral version info
# @author  Michael Hucka <mhucka@caltech.edu>
# @license Please see the file named LICENSE in the project directory
# @website https://github.com/casics/spiral
# =============================================================================

__version__     = '1.1.0'
__title__       = 'Spiral'
__description__ = 'SPlitters for IdentifieRs: A Library'
__url__         = 'https://github.com/casics/spiral'
__author__      = 'Michael Hucka <mhucka@caltech.edu>'
__email__       = 'casics-team@googlegroups.com'
__license__     = 'GNU General Public License, version 3.0'
__copyright__   = 'Copyright (C) 2015-2017 by the California Institute of Technology'
__citation__    = 'Hucka, (2018). Spiral: splitters for identifiers in source code files. Journal of Open Source Software, 3(24), 653, https://doi.org/10.21105/joss.00653'



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

from .code_chunker import CodeChunker
from .code_cleaner import CodeCleaner
from .code_pos import CodePOS
from .code_stopwords import NOUN_STOPWORDS, VERB_STOPWORDS
from .delimiter import Delimiter
from .lemmatizer import Lemmatizer
from .pair_checker import PairChecker
from .dictionary import *
from .code2vec import *
from .verb_helper import VerbHelper
from .dependency_analyzer import DependencyAnalyzer


'''__main__: main entry point for command-line interface

This module implements a command-line interface to Spiral.  It allows users
to invoke Spiral splitters from terminal shell command lines, to explore the
splitters or use them as part of toolchains.

Authors
-------

Michael Hucka <mhucka@caltech.edu>

Copyright
---------

Copyright (c) 2017 by the California Institute of Technology.  This software
was developed as part of the CASICS project, the Comprehensive and Automated
Software Inventory Creation System.  For more, visit http://casics.org.
'''

import os
import plac
import sys

import spiral
from spiral import ronin
from spiral import samurai
from spiral.simple_splitters import pure_camelcase_split, safe_simple_split
from spiral.simple_splitters import simple_split, elementary_split
from spiral.simple_splitters import heuristic_split


# Global constants.
# ......................................................................

_available_splitters = {'ronin': (ronin.init, ronin.split),
                        'samurai': (samurai.init, samurai.split),
                        'pure_camelcase_split': (None, pure_camelcase_split),
                        'safe_simple_split': (None, safe_simple_split),
                        'simple_split': (None, simple_split),
                        'elementary_split': (None, elementary_split),
                        'heuristic_split': (None, heuristic_split)}


# Main program.
# ......................................................................

@plac.annotations(
    file        = ('read input from a file',                     'option', 'f'),
    list        = ('list available splitters',                   'flag',   'l'),
    splitter    = ('run named splitter, e.g. "ronin"',           'option', 's'),
    version     = ('print version info and exit',                'flag',   'V'),
    strings     = 'text string(s) to split',
)

def main(list=False, file=None, splitter='ronin', version=False, *strings):
    '''Spiral is "SPlitters for IdentifieRs: A Library".  It implements many
identifier splitting algorithms.  This command-line program provides a very
simple interface to run Spiral's splitters, for testing and exploration.

IMPORTANT: this is NOT meant to be the normal interface to using Spiral.
Spiral is a package meant to be used by other Python programs.  When you run
this command-line program, it needs to load a large data file internally
and this will cause a noticeable start-up delay.  In normal usage, inside an
application program, Spiral would only load the data files once at first
invocation and subsequent calls to Spiral splitters would not be so slow.
However, since this command-line program cannot save the internal data across
invocations, the startup cost occurs every time.  Please rest assured that
this startup delay is only a one-time cost and not typical for normal Spiral
usage.

The input to this program can be one or more strings on the command line,
or (using the -f argument) a file of strings.  If given a file of
strings, it will analyze each line in the file separately.

The splitter to be used can be specified using the -s option.  Use the -l
option to print a list of the splitters available.

The optional argument -v will make this program display version information
and exit without doing anything more.
'''
    # Process arguments
    if version:
        print('{} version {}'.format(spiral.__title__, spiral.__version__))
        print('Author: {}'.format(spiral.__author__))
        print('URL: {}'.format(spiral.__url__))
        print('License: {}'.format(spiral.__license__))
        sys.exit()
    if list:
        print('Available splitters:')
        for name, _ in sorted(_available_splitters.items()):
            print('  ' + name)
        sys.exit()
    if splitter and splitter not in _available_splitters:
        raise SystemExit('Unrecognized splitter: "{}"'.format(splitter))
    if not file and not strings:
        raise SystemExit('Need a file or string as input. Use -h to get help.')
    if strings and strings[0].startswith('-'):
        # If it starts with a dash and we get to this point, it's not an arg
        # recognized by plac and it's probably not a string meant to be split.
        raise SystemExit('Unrecognized argument "{}". (Hint: use -h to get help.)'
                         .format(strings[0]))

    # Let's do this thing.
    if file:
        if os.path.exists(file):
            with open(file) as f:
                split(f.readlines(), _available_splitters[splitter])
        elif os.path.exists(os.path.join(os.getcwd(), file)):
            with open(os.path.join(os.getcwd(), file)) as f:
                split(f.readlines(), _available_splitters[splitter])
        else:
            raise ValueError('Cannot find file "{}"'.format(file))
    else:
        split(strings, _available_splitters[splitter])


def split(string_list, splitter_functions):
    init, split = splitter_functions
    if init:
        init()
    for s in [string.rstrip() for string in string_list]:
        print('{}: {}'.format(s, split(s)))


if __name__ == '__main__':
    plac.call(main)



# For Emacs users
# ......................................................................
# Local Variables:
# mode: python
# python-indent-offset: 4
# End:



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import itertools

from .javalang.parse import parse as ast_parse
from .javalang.tokenizer import Literal
from .javalang.tree import CompilationUnit, ClassDeclaration, FieldDeclaration, MemberReference, MethodDeclaration, MethodInvocation
from .delimiter import Delimiter
from .code_pos import CodePOS
from .code_chunker import CodeChunker
from .lemmatizer import Lemmatizer
from .dictionary import PREPOSITIONS


class ActionParser:
    __INSTANCE = None

    def __init__(
        self,
        code_pos=CodePOS.get_inst(),
        code_chunker=CodeChunker.get_inst(),
    ):
        self.code_pos = code_pos
        self.code_chunker = code_chunker

    @classmethod
    def get_inst(cls, *args):
        if cls.__INSTANCE is None:
            cls.__INSTANCE = cls()
        return cls.__INSTANCE


    def __call__(self, *args):
        return self.parse(*args)

    def parse_method_declaration(self, clazz_name, method_name, params, ret):
        verb2nameroles = self.parse_method_name(method_name, ret)
        verb2roles = self.complete_role(verb2nameroles, clazz_name, params)
        return verb2roles

    def parse_method_name(self, name, ret):
        segmented_name = Delimiter.split_camel(name)
        pos = self.code_pos(segmented_name)
        verbs = {word for word, tag in pos if tag ==
                 "verb" and not word.endswith("ed") and not word.endswith("ing")}
        verb2chunks = self.code_chunker(segmented_name, *verbs)

        verb2nameroles = {}
        for verb, chunks in verb2chunks.items():
            if verb is None:
                continue
            lemma = Lemmatizer.lemmatize_verb(verb)
            name_role = []
            verb_index = -1
            prep_indices = []
            for index, chunk in enumerate(chunks):
                if chunk == verb:
                    verb_index = index
                elif chunk in PREPOSITIONS and verb_index >= 0 and index > verb_index:
                    prep_indices.append(index)
            if verb_index > 0:
                prefix = " ".join(chunks[:verb_index])
                if ret.lower() == "boolean":
                    name_role.append((prefix, "subj"))
                else:
                    # name_role.append((prefix, "advmod"))
                    verb = f"{prefix} {verb}"
            name_role.append((verb, "verb"))

            if len(prep_indices) > 0:
                if prep_indices[0] > verb_index + 1:
                    dobj = " ".join(chunks[verb_index + 1:prep_indices[0]])
                    name_role.append((dobj, "dobj"))
                for start, end in zip(prep_indices, prep_indices[1:] + [len(chunks)]):
                    prep = chunks[start]
                    name_role.append((prep, "prep"))
                    if start + 1 < end:
                        pobj = " ".join(chunks[start + 1:end])
                        name_role.append((pobj, "pobj"))
                    else:
                        pobj = None
                    # iobj.append((prep, pobj))
            else:
                if verb_index + 1 < len(chunks):
                    dobj = " ".join(chunks[verb_index + 1:])
                    name_role.append((dobj, "dobj"))
            verb2nameroles[verb] = name_role
        return verb2nameroles

    def complete_role(self, verb2nameroles: dict, clazz_name, params):
        verb2roles = dict()
        segmented_clazz = Delimiter.split_camel(clazz_name)
        for verb, name_roles in verb2nameroles.items():
            name_nps = {phrase for phrase,
                        role in name_roles if role not in {"verb", "prep"}}
            params = [Delimiter.split_camel(param) for param in params]
            params = [param for param in params if param not in name_nps]
            size = len(name_roles)

            phrases, roles = [], []
            for index, (phrase, role) in enumerate(name_roles):
                phrases.append(phrase)
                roles.append(role)
                if role == "verb":
                    if index + 1 < size and name_roles[index + 1][1] == "prep":
                        phrases.append(segmented_clazz)
                        roles.append("dobj")
                    elif index + 1 == size:
                        if len(params) > 0:
                            phrases.append(params.pop(0))
                            roles.append("dobj")
                        else:
                            phrases.append(segmented_clazz)
                            roles.append("dobj")
                elif role == "prep" and index + 1 < size and name_roles[index + 1][1] != "pobj" and len(params) > 0:
                    phrases.append(params.pop(0))
                    roles.append("pobj")

            if len(roles) > 0 and roles[-1] == "prep" and len(params) > 0:
                phrases.append(params.pop(0))
                roles.append("pobj")

            # if len(roles) > 0 and roles[-1] != "pobj":
            #     lemma = Lemmatizer.lemmatize_verb(verb)
            #     preps = VERB_PATTERN.get(lemma, [])

            #     # TODO: improve this part
            #     if len(preps) > 0:
            #         prep = preps[0]
            #         if len(params) > 0:
            #             phrases.append(prep)
            #             roles.append("prep")
            #             phrases.append(params.pop(0))
            #             roles.append("pobj")
            #         else:
            #             phrases.append(prep)
            #             roles.append("prep")
            #             phrases.append(segmented_clazz)
            #             roles.append("pobj")

            subj = None
            dobj = None
            iobj = []
            size = len(phrases)
            index = 0
            while index < size:
                phrase = phrases[index]
                role_type = roles[index]
                if role_type == "subj":
                    subj = phrase
                elif role_type == "dobj":
                    dobj = phrase
                elif role_type == "prep" and index + 1 < size:
                    iobj.append((phrase, phrases[index + 1]))
                    index += 1
                index += 1

            verb2roles[verb] = dict()
            if subj is not None:
                verb2roles[verb]["subj"] = subj
            verb2roles[verb]["verb"] = verb
            if dobj is not None:
                verb2roles[verb]["dobj"] = dobj
            if len(iobj) > 0:
                verb2roles[verb]["iobj"] = iobj
        return verb2roles

    #处理参数中的MethodInvocation
    def deal_invocation(self, methodInvocation):
        scope = str(methodInvocation.qualifier)
        invoke_method = str(methodInvocation.member)
        arguments = methodInvocation.arguments
        arg_list = list()
        for arg in arguments:
            if isinstance(arg, MemberReference):
                if arg.qualifier is None:
                    arg_list.append(str(arg.member))
                else:
                    arg_list.append(arg.qualifier + arg.member)
            if isinstance(arg, Literal):
                arg_list.append(arg.value)
            if isinstance(arg, MethodInvocation):
                arg_list.append(self.deal_invocation(arg))
        str_invocation = scope.join(".").join(invoke_method).join("(")
        for item in arg_list:
            str_invocation.join(item)
        str_invocation.join(")")
        return str_invocation

    #把生成的主谓宾组成三元组
    def convert_to_triples(self, action, method_name : str):
        desc = list()
        for key in action.keys():
            content = action[key]
            #subj不存在则把方法名当作subj
            if 'subj' in content:
                subj = content['subj']
            else:
                subj = method_name

            verb = content['verb']
            dobj = content['dobj']
            
            if 'iobj' in content:
                iobjs = content['iobj']
            else:
                iobjs = None

            sub_desc = (subj, verb, dobj)
            desc.append(sub_desc)
            if iobjs is not None:
                for iobj in iobjs:
                    temp = str("")
                    for word in iobj:
                        temp = temp + str(word) + " "
                    temp = temp[:-1]
                    sub_desc = (subj, verb, temp)
                    desc.append(sub_desc)
        return desc

    #提取代码中的方法调用，并且生成相应的句子成分
    def parse(self, code):
        if isinstance(code, CompilationUnit):
            tree = code
        else:
            try:
                tree = ast_parse(code)
            except Exception:
                return []
        clazz_decl = (tree.types)[0]
        clazz_name = clazz_decl.name
        #获取class中的method
        for node in clazz_decl.body:
            if isinstance(node, MethodDeclaration):
                method_decl = node
        #获取method的返回值
        if method_decl.return_type is None:
            ret = "void"
        else:
            ret = method_decl.return_type

        method_params = list()
        for param in method_decl.parameters:
            method_params.append(param.type.name)
            method_params.append(param.name)
        #主谓宾结果集
        results = list()

        #首先将方法声明主谓宾加入结果集
        action = self.parse_method_declaration(clazz_name, method_decl.name, method_params, ret)
        result = self.convert_to_triples(action, method_decl.name)[0]
        results.append(result)
        #获取method中的MethodInvocation，并生成相应主谓宾加入结果集
        for _, node in method_decl:
            if isinstance(node, MethodInvocation):
                scope = node.qualifier
                invoke_method = node.member
                invoke_arguments = list()
                for arg in node.arguments:
                    if isinstance(arg, MemberReference):
                        if arg.qualifier is None:
                            invoke_arguments.append(str(arg.member))
                        else:
                            invoke_arguments.append(arg.qualifier + arg.member)
                    if isinstance(arg, MethodInvocation):
                        invoke_arguments.append(self.deal_invocation(arg))                        
                    if isinstance(arg, Literal):
                        invoke_arguments.append(str(arg.value))
                action = self.parse_method_declaration(scope, invoke_method, invoke_arguments, ret)
                desc = self.convert_to_triples(action, invoke_method)
                if len(desc) > 1:
                    for item in desc:
                        results.append(item)
                else:
                    results.append(desc[0])
        return results


#!/usr/bin/env python
# -*- encoding: utf-8 -*-

from pathlib import Path
import pickle
import math
import time
import multiprocessing
import logging
import itertools

import numpy as np
from gensim.models.fasttext import FastTextKeyedVectors

from codetoolkit.code2vec import  _extract_for_code, generate_corpus, train


if __name__ == "__main__":

    # with Path("test/Sample1.java").open("r") as f:
    #     code  = f.read()
    # seqs, vocab = _extract_for_code(code, level="class")
    # print(vocab)
    # for seq in seqs:
    #     print(seq)

    # format = '[%(levelname)s] %(asctime)s - %(pathname)s[line:%(lineno)d] - %(message)s'
    # logging.basicConfig(level=logging.INFO, format=format)
    # formatter = logging.Formatter(format)
    # file_handler = logging.FileHandler(filename="test/code2vec-classlevel.log", mode="w", encoding='utf-8')
    # file_handler.setFormatter(formatter)
    # file_handler.setLevel(logging.INFO)
    # logging.getLogger().addHandler(file_handler)
            

    # inp_files = list()
    # seq_files = list()
    # meta_files = list()
    # for file in Path("/home/Data/SemanticTagging/codebase").glob("classes-batch*"):
    #     batch_id = file.parts[-1].split(".")[-2][len("classes-batch"):]
    #     inp_files.append(str(file))
    #     seq_files.append(f"/home/Data/CodeToolkit/code2vec/sequence_corpus-classlevel-batch{batch_id}.txt")
    #     meta_files.append(f"/home/Data/CodeToolkit/code2vec/sequence_corpus_meta-classlevel-batch{batch_id}.pkl")
    # print(inp_files)

    # start_time = time.time()
    # pool = multiprocessing.Pool(len(inp_files))
    # results = []
    # for input_file, output_file, vocab_file in zip(inp_files, seq_files, meta_files):
    #     rs = pool.apply_async(generate_corpus, args=(input_file, output_file, vocab_file, "class"))
    #     results.append(rs)
    # pool.close()
    # pool.join()
    # print(f"generate corpus time: {time.time() - start_time}s")

    # start_time = time.time()
    # train(seq_files, meta_files, model_path="/home/Data/CodeToolkit/code2vec/fasttext-classlevel.bin", emb_path="/home/Data/CodeToolkit/code2vec/emb-classlevel.bin", fine_tuning=True, epochs=2)
    # print(f"train token vectors: {time.time() - start_time}s")




    # with Path(f"/home/Data/CodeToolkit/code2vec/sequence_corpus-batch1.txt").open("r", encoding="utf-8") as f:
    #     for i, line in enumerate(f):
    #         print(line.strip())
    #         if i >= 10000:
    #             break
        

    def cos(code2vec, word1, word2):
        words1 = word1.split()
        words2 = word2.split()
        vecs1 = [code2vec.get_vector(w) for w in words1]
        vecs2 = [code2vec.get_vector(w) for w in words2]
        vec1 = sum(vecs1) / len(vecs1)
        vec2 = sum(vecs2) / len(vecs2)
        # print(sum(vec1), sum(vec2))
        cos = np.dot(vec1, vec2) / np.linalg.norm(vec1) / np.linalg.norm(vec2)
        return cos / 2 + 0.5

    def pair_cos(code2vec, pair1, pair2):
        vec1 = code2vec.get_vector(pair1[0]) - code2vec.get_vector(pair1[1])
        vec2 = code2vec.get_vector(pair2[0]) - code2vec.get_vector(pair2[1])
        # print(sum(vec1), sum(vec2))
        cos = np.dot(vec1, vec2) / np.linalg.norm(vec1) / np.linalg.norm(vec2)
        return cos / 2 + 0.5

    code2vec: FastTextKeyedVectors = FastTextKeyedVectors.load("/home/Data/ResOpMining/CodeSearchData/code_search_emb-methodlevel.bin")

    # print(code2vec.most_similar("md5sum", topn=100))
    # print(code2vec.most_similar("md5", topn=100))
    # print(code2vec.most_similar("db", topn=100))

    # print(code2vec.most_similar("row", topn=100))
    # print(code2vec.most_similar("col", topn=100))

    # print(code2vec.most_similar("statement", topn=100))
    # print(code2vec.most_similar("stmt", topn=100))

    # keywords = {('input', 'input'), ('file', 'files'), ('message', 'message'), ('digest', 'digest'), ('input stream', 'input stream'), ('message digest', 'message digest'), ('md5', 'md5'), ('path', 'paths')}
    # for (w1, _), (w2, _) in itertools.combinations(keywords, 2):
    #     print(w1, w2, cos(code2vec, w1, w2))
    
    # print("database" in code2vec)
    # print("luminance" in code2vec)
    # print("lum" in code2vec)
    # print("md5" in code2vec)
    # for word, sim in code2vec.most_similar("md5sum", topn=100000):
    #     print(word, sim)
    # print(cos(code2vec, "clear", "reset"))
    # print(cos(code2vec, "substr", "substring"))
    # print(cos(code2vec, "rows", "columns"))
    # print(cos(code2vec, "db", "database"))
    # print(cos(code2vec, "md5", "file"))
    
    # print(cos(code2vec, "rr", "repfdsafsda"))
    # print(cos(code2vec, "x", "y"))

    print(pair_cos(code2vec, ['lock', 'acquire'], ['unlock', 'release']))


from codetoolkit.dependency_analyzer import DependencyAnalyzer

if __name__ == '__main__':
    code  = open("test/Sample1.java", "r").read()
    analyzer = DependencyAnalyzer()
    cfg = analyzer.build_cfg(code)
    pdg = analyzer.build_pdg(code)
    print(cfg)
    print(pdg)




'''
frequencies: code to handle word frequencies
'''


# Main code.
# .............................................................................

def frequencies_from_csv_file(filename, threshold=None, filter_words=None):
    '''Read a table of frequencies (as a Python dictionary) from a CSV file.
    Parameter 'threshold' sets a minimum frequency value for entries to keep.
    Entries with frequency values less than the given 'threshold' value will
    be discarded.  Parameter 'filter_word' is a set of words (dictionary keys)
    that should be ignored.  Entries that are found in 'filter_words' will be
    discarded.  Returns a dictionary of the results.
    '''
    import csv
    try:
        with open(filename, 'r') as f:
            reader = csv.DictReader(f, fieldnames=['word','frequency'])
            frequencies = {}
            for row in reader:
                word = row['word']
                frequency = int(row['frequency'])
                if threshold and frequency < threshold:
                    continue
                if filter_words and word in filter_words:
                    continue
                frequencies[word] = frequency
            return frequencies
    except Exception as err:
        return {}


def frequencies_from_pickle(filename, threshold=None, filter_words=None):
    '''Read a table of frequencies (as a Python dictionary) from a pickle file.
    If the file name ends in the extension .pklz, it's assumed to be a gzip'ed
    pickle file; otherwise, it's assumed to be an uncompressed pickle file.
    Parameter 'threshold' sets a minimum frequency value for entries to keep.
    Entries with frequency values less than the given 'threshold' value will
    be discarded.  Parameter 'filter_word' is a set of words (dictionary keys)
    that should be ignored.  Entries that are found in 'filter_words' will be
    discarded.  Returns a dictionary of the results.
    '''
    import pickle
    data = {}
    try:
        if filename.endswith('.pklz'):
            import gzip
            with gzip.open(filename, 'rb') as pickle_file:
                data = pickle.load(pickle_file)
        else:
            with open(filename, 'rb') as pickle_file:
                data = pickle.load(pickle_file)
        if threshold:
            data = {k : v for k, v in data if v < threshold}
        if filter_words:
            data = {k : v for k, v in data if k not in filter_words}
        return data
    except Exception as err:
        return {}


def save_frequencies_to_pickle(data, filename):
    '''Write data to a pickle file.  If the file ends in the extension .pklz,
    it's assumed to be a gzip'ed pickle file; otherwise, it's assumed to be
    an uncompressed pickle file.
    '''
    import pickle
    try:
        if filename.endswith('.pklz'):
            import gzip
            with gzip.open(filename, 'wb') as pickle_file:
                pickle.dump(data, pickle_file)
        else:
            with open(filename, 'wb') as pickle_file:
                pickle.dump(data, pickle_file)
    except IOError as err:
        raise SystemExit('encountered error trying to dump pickle {}'.format(filename))
    except pickle.PickleError as err:
        raise SystemExit('pickling error for {}'.format(filename))



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import functools
import nltk
from nltk.stem import WordNetLemmatizer
from nltk.corpus import wordnet

LEMMATIZER = WordNetLemmatizer()

class Lemmatizer:
    @functools.lru_cache(maxsize=10000)
    def lemmatize_noun(noun):
        lower = noun.lower()
        return LEMMATIZER.lemmatize(lower, "n")

    @functools.lru_cache(maxsize=10000)
    def lemmatize_np(np):
        return " ".join([Lemmatizer.lemmatize_noun(word) for word in np.split()])

    @functools.lru_cache(maxsize=10000)
    def lemmatize_verb(verb):
        lower = verb.lower()
        words = lower.split()
        lemma = " ".join([LEMMATIZER.lemmatize(words[0], "v")] + words[1:])
        return lemma

    @functools.lru_cache(maxsize=10000)
    def check_verb(word, rate_thres=0.6, count_thres=10):
        pos2count = dict()
        for synset in wordnet.synsets(word):
            lemma, pos, _ = synset.name().split(".")
            pos2count[pos] = pos2count.get(pos, 0) + 1
        # logging.info(pos2count)
        if len(pos2count) > 0 and (pos2count.get("v", 0) >= count_thres or pos2count.get("v", 0) / sum(pos2count.values()) >= rate_thres):
            return True
        return False

    # @functools.lru_cache(maxsize=10000)
    # def check_noun(word, rate_thres=0.5):
    #     pos2count = dict()
    #     for synset in wordnet.synsets(word):
    #         parts = synset.name().split(".")
    #         if len(parts) != 3:
    #             continue
    #         lemma, pos, _ = parts
    #         pos2count[pos] = pos2count.get(pos, 0) + 1
    #     # logging.info(pos2count)
    #     if len(pos2count) == 0:
    #         return True
    #     if pos2count.get("n", 0) / sum(pos2count.values()) >= rate_thres:
    #         return True
    #     return False

    @functools.lru_cache(maxsize=10000)
    def check_noun(word):
        # logging.info(nltk.pos_tag([word]))
        return nltk.pos_tag([word])[0][1] in {"NN", "NNS"}

    # @functools.lru_cache(maxsize=10000)
    # def check_noun(word, rate_thres=0.5):
    #     pos2count = dict()
    #     for synset in wordnet.synsets(word):
    #         parts = synset.name().split(".")
    #         if len(parts) != 3:
    #             continue
    #         lemma, pos, _ = parts
    #         pos2count[pos] = pos2count.get(pos, 0) + 1
    #     # logging.info(pos2count)
    #     if len(pos2count) == 0:
    #         return True
    #     if pos2count.get("n", 0) / sum(pos2count.values()) >= rate_thres:
    #         return True
    #     return False


import logging
logging.basicConfig(level = logging.DEBUG)

from codetoolkit.delimiter import Delimiter

if __name__ == "__main__":

    print(Delimiter.split_camel("codepoint"))
    print(Delimiter.split_camel("makelayer"))
    print(Delimiter.split_camel("dealwithread"))


from pathlib import Path

class VerbHelper:
    __INSTANCE = None
    VERB_TENSES_KEYS = {
        "infinitive"           : 0,
        "1st singular present" : 1,
        "2nd singular present" : 2,
        "3rd singular present" : 3,
        "present plural"       : 4,
        "present participle"   : 5,
        "1st singular past"    : 6,
        "2nd singular past"    : 7,
        "3rd singular past"    : 8,
        "past plural"          : 9,
        "past"                 : 10,
        "past participle"      : 11
    }

    VERB_TENSES_ALIASES = {
        "inf"     : "infinitive",
        "1sgpres" : "1st singular present",
        "2sgpres" : "2nd singular present",
        "3sgpres" : "3rd singular present",
        "pl"      : "present plural",
        "prog"    : "present participle",
        "1sgpast" : "1st singular past",
        "2sgpast" : "2nd singular past",
        "3sgpast" : "3rd singular past",
        "pastpl"  : "past plural",
        "ppart"   : "past participle"
    }

    def __init__(self, dictionary=str(Path(__file__).parent / "verb.txt")):
        self.verb_tenses = {}
        with Path(dictionary).open("r", encoding="utf-8") as f:
            for line in f:
                strs = line.strip().split(",")
                self.verb_tenses[strs[0]] = strs

        self.verb_lemmas = {}
        for infinitive in self.verb_tenses:
            for tense in self.verb_tenses[infinitive]:
                if tense != "":
                    self.verb_lemmas[tense] = infinitive

    @classmethod
    def get_inst(cls, *args):
        if cls.__INSTANCE is None:
            cls.__INSTANCE = cls(*args)
        return cls.__INSTANCE

    def verb_conjugate(self, verb, tense="infinitive", negate=False):
        """Inflects the verb to the given tense.
        For example: be
        present: I am, you are, she is,
        present participle: being,
        past: I was, you were, he was,
        past participle: been,
        negated present: I am not, you aren't, it isn't.
        """
        _verb = self.verb_lemmas.get(verb, "")
        i = VerbHelper.VERB_TENSES_KEYS[tense]
        if negate is True: 
            i += len(VerbHelper.VERB_TENSES_KEYS)
        if _verb not in self.verb_tenses:
            return verb
        return self.verb_tenses[_verb][i]

    def verb_past(self, verb, negate=False):
        elements = verb.split()
        _verb = elements[-1]
        
        if _verb not in self.verb_lemmas:
            if _verb.endswith("e"):
                _verb = f"{_verb}d"
            else:
                _verb = f"{_verb}ed"
            adv = elements[:-1]
            return " ".join(adv + [_verb])

        _verb = self.verb_lemmas[_verb]
        index = VerbHelper.VERB_TENSES_KEYS["past"]
        if negate is True: 
            index += len(VerbHelper.VERB_TENSES_KEYS)

        if _verb not in self.verb_tenses or self.verb_tenses[_verb][index] == "":
            if _verb.endswith("e"):
                _verb = f"{_verb}d"
            else:
                _verb = f"{_verb}ed"
            adv = elements[:-1]
            return " ".join(adv + [_verb])

        return " ".join(elements[:-1] + [self.verb_tenses[_verb][index]])


# =============================================================================
# @file    data_utils.py
# @brief   CASICS Spiral internal utilities for dealing with data
# @author  Michael Hucka <mhucka@caltech.edu>
# @license Please see the file named LICENSE in the project directory
# @website https://github.com/casics/spiral
# =============================================================================

import collections

def msg(string, *other_args):
    '''Like the standard print(), but treats the first argument as a string
    with format specifiers, and also flushes the output immediately. Flushing
    immediately is useful when piping the output of a script, because Python
    by default will buffer the output in that situation and this makes it
    very difficult to see what is happening in real time.
    '''
    print(string.format(*other_args), flush=True)


def flatten(lst):
    '''Recursively flatten a list of lists.  From the answer by user Christian
    here: https://stackoverflow.com/a/2158532/743730'''
    for el in lst:
        if isinstance(el, collections.Iterable) and not isinstance(el, (str, bytes)):
            yield from flatten(el)
        else:
            yield el



'''simple_splitters: simple identifier splitters

This exports a number of simple splitter functions whose behaviors are
sightly different depending on what assumptions are made about identifier
patterns.  All are simple in the sense that they do not make complicated
inferences about the string and at most (in the cause of heuristic_split) use
fixed dictionaries of predefined terms to enhance performance.

The following table illustrates the differences between the different splitters
implemented in this file.  The first column shows the input string; the
remaining columns show the results of the different X_split() functions.

input        pure_camel    safe_simple  simple        elementary     heuristic
-------      -------       -------      -------       -------        -------
alllower     alllower      alllower     alllower      alllower       alllower
ALLUPPER     ALLUPPER      ALLUPPER     ALLUPPER      ALLUPPER       ALLUPPER
a_delimiter  a_delimiter   a delimiter  a delimiter   a delimiter    a delimiter
a.delimiter  a.delimiter   a delimiter  a delimiter   a delimiter    a delimiter
a$delimiter  a$delimiter   a delimiter  a delimiter   a delimiter    a delimiter
a:delimiter  a:delimiter   a delimiter  a delimiter   a delimiter    a delimiter
a_fooBar     a_foo Bar     a foo Bar    a foo Bar     a foo Bar      a foo Bar
fooBar       foo Bar       foo Bar      foo Bar       foo Bar        foo Bar
FooBar       Foo Bar       Foo Bar      Foo Bar       Foo Bar        Foo Bar
Foobar       Foobar        Foobar       Foobar        Foobar         Foobar
fooBAR       foo BAR       fooBAR       foo BAR       foo BAR        foo BAR
fooBARbif    foo BARbif    fooBARbif    foo BARbif    foo BARbif     foo BARbif
fooBARzBif   foo BARz Bif  fooBARzBif   foo BARz Bif  foo BARz Bif   foo BARz Bif
ABCfoo       ABCfoo        ABCfoo       ABCfoo        ABCfoo         ABCfoo
ABCFoo       ABCFoo        ABCFoo       ABCFoo        ABCFoo         ABCFoo
ABCFooBar    ABCFoo Bar    ABCFooBar    ABCFoo Bar    ABCFoo Bar     ABCFoo Bar
ABCfooBar    ABCfoo Bar    ABCfooBar    ABCfoo Bar    ABCfoo Bar     ABCfoo Bar
fooBar2day   foo Bar2day   foo Bar2day  foo Bar2day   foo Bar 2 day  foo Bar 2 day
fooBar2Day   foo Bar2Day   foo Bar2Day  foo Bar2Day   foo Bar 2 Day  foo Bar 2 Day
fooBAR2day   foo BAR2day   fooBAR2day   foo BAR2day   foo BAR 2 day  foo BAR 2 day
fooBAR2Day   foo BAR2Day   fooBAR2Day   foo BAR2Day   foo BAR 2 Day  foo BAR 2 Day
foo3000      foo3000       foo3000      foo3000       foo 3000       foo 3000
99foo3000    99foo3000     99foo3000    99foo3000     99 foo 3000    99 foo 3000
foo2Bar      foo2Bar       foo2Bar      foo2Bar       foo 2 Bar      foo 2 Bar
foo2bar2     foo2bar2      foo2bar2     foo2bar2      foo 2 bar 2    foo 2 bar 2
Foo2Bar2     Foo2Bar2      Foo2Bar2     Foo2Bar2      Foo 2 Bar 2    Foo 2 Bar 2
MyInt32      My Int32      My Int32     My Int32      My Int 32      My Int32
MyInt42      My Int42      My Int42     My Int42      My Int 42      My Int 42
MyInt32Var   My Int32Var   My Int32Var  My Int32Var   My Int 32 Var  My Int32 Var
2ndvar       2ndvar        2ndvar       2ndvar        2 ndvar        2nd var
the2ndvar    the2ndvar     the2ndvar    the2ndvar     the 2 ndvar    the 2nd var
the2ndVar    the2nd Var    the2nd Var   the2nd Var    the 2 nd Var   the 2nd Var
row10        row10         row10        row10         row 10         row 10
utf8         utf8          utf8         utf8          utf 8          utf8
aUTF8var     a UTF8var     aUTF8var     a UTF8var     a UTF 8 var    a UTF8 var
J2SE4me      J2SE4me       J2SE4me      J2SE4me       J 2 SE 4 me    J2SE 4 me
IPv4addr     IPv4addr      IPv4addr     IPv4addr      IPv 4 addr     IPv4 addr
'''

import re
import string
import sys

try:
    sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
    sys.path.append(os.path.join(os.path.dirname(__file__), "../spiral"))
except:
    sys.path.append("..")
    sys.path.append("../spiral")

try:
    from .constants import *
    from .utils import flatten
except:
    from constants import *
    from utils import flatten


# Global definitions.
# .............................................................................

_delimiter_chars = '$~_.:/@'
_two_capitals    = re.compile(r'[A-Z][A-Z]')
_camel_case      = re.compile(r'((?<=[a-z])[A-Z])')


# Delimiter-based splitter.
# .............................................................................
#
# This does nothing fancy. It splits by explicit delimiter characters like '_'.

_delimiter_splitter = str.maketrans(_delimiter_chars, ' '*len(_delimiter_chars))

def delimiter_split(identifier):
    '''Split identifier by explicit delimiters only.'''
    parts = str.translate(identifier, _delimiter_splitter).split(' ')
    parts = [p for p in parts if p]
    return parts


_digit_splitter = str.maketrans(string.digits, ' '*len(string.digits))

def digit_split(identifier):
    '''Split identifier at digits only.'''
    parts = str.translate(identifier, _digit_splitter).split(' ')
    parts = [p for p in parts if p]
    return parts


# Pure camel case splitter.
# .............................................................................

def pure_camelcase_split(identifier, safe=False):
    '''Split identifiers by forward camel case only, i.e., lower-to-upper
    case transitions.  This means it will split fooBarBaz into 'foo', 'Bar'
    and 'Baz', but it won't change SQLlite or similar identifiers. It also
    ignores delimiter characters.  It will not split identifies that have
    multiple adjacent uppercase letters unless parameter 'safe' is True; for
    example, 'setAVariable' -> ['set', 'AVariable'] with the default value of
    safe == False, but 'setAVariable' -> ['setAVariable'] if safe == True.
    (The rationale is "better safe than sorry" for cases that cannot be split
    without a dictionary or heuristics.)
    '''
    if safe and re.search(_two_capitals, identifier):
        return [identifier]
    return re.sub(_camel_case, r' \1', identifier).split()


# Safe simple splitter.
# .............................................................................

_hard_splitter = str.maketrans(_delimiter_chars, ' '*len(_delimiter_chars))

def safe_simple_split(identifier):
    '''Split identifiers by hard delimiters ('_', '$', '~', '.', ':') and
    forward camel case only, i.e., lower-to-upper case transitions.  This
    means it will split fooBarBaz into 'foo', 'Bar' and 'Baz', but it won't
    split SQLlite or similar identifiers.  Does not split identifies that
    have multiple adjacent uppercase letters anywhere in them, because doing
    so is risky if the uppercase letters are not an acronym.  Example:
    aFastNDecoder -> ['aFastNDecoder'].  Digits are not removed from strings
    nor treated as delimiters, although a capital letter after a digit is
    considered equivalent to a camel case transition; for example,
    'foo_bar2Biff' will be split as ['foo', 'bar2', 'Biff'].
    '''
    parts = str.translate(identifier, _hard_splitter).split()
    return list(flatten(pure_camelcase_split(token, safe=True) for token in parts))


# Not-so-safe simple splitter.
# .............................................................................

def simple_split(identifier):
    '''Split identifiers by hard delimiters such as underscores, and forward
    camel case only, i.e., lower-to-upper case transitions.  This means it
    will split fooBarBaz into 'foo', 'Bar' and 'Baz', and foo2bar into 'foo'
    and 'bar, but it won't change SQLlite or similar identifiers.  Unlike
    safe_simple_split(), this will split identifiers that may have sequences
    of all upper-case letters if there is a lower-to-upper case transition
    somewhere.  Example: simple_split('aFastNDecoder') will produce ['a',
    'Fast', 'NDecoder'] even though "NDecoder" may be more correctly split as
    'N' 'Decoder'.  It preserves digits and does not treat them specially.
    '''
    parts = str.translate(identifier, _hard_splitter).split()
    return list(flatten(re.sub(_camel_case, r' \1', token).split() for token in parts))


# Not-so-safe simple splitter.
# .............................................................................

def elementary_split(identifier):
    '''Split identifiers by hard delimiters such as underscores, digits, and
    forward camel case only, i.e., lower-to-upper case transitions.  This
    means it will split fooBarBaz into 'foo', 'Bar' and 'Baz', and foo2bar
    into 'foo' and 'bar, but it won't change SQLlite or similar identifiers.
    Unlike safe_simple_split(), this will split identifiers that may have
    sequences of all upper-case letters if there is a lower-to-upper case
    transition somewhere.  Example: simple_split('aFastNDecoder') will
    produce ['a', 'Fast', 'NDecoder'] even though "NDecoder" may be more
    correctly split as 'N' 'Decoder'.  Digits are treated as delimiters, but
    not otherwise recognized or removed.  For example, 'utf8_var' will be
    split into ['utf', '8', 'var'].  (Contrast this to heuristic_split().)
    '''
    split_str = str.translate(identifier, _hard_splitter)
    split_str = re.sub(r'(\d+)', r' \1 ', split_str)
    return re.sub(_camel_case, r' \1', split_str).split()


# Not-so-safe, not-so-simple splitter.
# .............................................................................

_exceptions_re = re.compile(r'(' + '|'.join(sorted(common_terms_with_numbers, key=lambda term: len(term), reverse=True)) + ')', re.I)
_common_suffixes = tuple(common_suffix_numbers)

# This next pattern matches hex numbers, octal numbers, and other numbers, in
# that order. Order is important, or else we'd split hex numbers.
_numbers_re = re.compile(r'(0x[0-9a-f]+l?|0[1-7]{2,}|\d+)', re.I)

def heuristic_split(identifier, keep_numbers=True, exceptions=None):
    '''Split identifiers by hard delimiters such as underscores, digits, and
    forward camel case only, i.e., lower-to-upper case transitions.  This
    means it will split fooBarBaz into 'foo', 'Bar' and 'Baz', and foo2bar
    into 'foo' and 'bar, but it won't change SQLlite or similar identifiers.
    Unlike safe_simple_split(), this will split identifiers that may have
    sequences of all upper-case letters if there is a lower-to-upper case
    transition somewhere.  Example: heuristic_split('aFastNDecoder') will
    produce ['a', 'Fast', 'NDecoder'] even though "NDecoder" may be more
    correctly split as 'N' 'Decoder'.  If keep_numbers=False, leading digits
    are removed from the identifier string before processing; embedded digits
    are removed if they appear at the tail ends of tokens, but not if they
    appear in the middle of tokens or are recognized as being exceptions
    defined by the set in parameter 'exceptions'.  (The default set of
    exceptions are common terms with embedded numbers such as 'utf8'.)
    Tokens that consist ONLY of digits are removed.
    '''
    if exceptions:
        exceptions_re = re.compile(r'(' + '|'.join(exceptions) + ')', re.I)
    else:
        exceptions = common_terms_with_numbers
        exceptions_re = _exceptions_re

    preliminary = str.translate(identifier, _hard_splitter)
    preliminary = re.sub(exceptions_re, r' \1 ', preliminary)
    preliminary = re.sub(_camel_case, r' \1', preliminary)
    if keep_numbers:
        # Unfortunately gnarly and hard-to-read expression.  Did this for the
        # sake of speed.  Basic idea: keep each token that is in the exceptions
        # list or ends with a common suffix; otherwise, split the token by digits
        # and clean up the resulting digit strings.
        parts = flatten(s if (_exceptions_re.match(s)
                              or (not s[0].isdigit() and s.endswith(_common_suffixes)))
                        else filter(bool, _numbers_re.split(s))
                        for s in preliminary.split())
    else:
        # Variation of the previous thing that will remove leading and trailing
        # digits of tokens, possibly removing the tokens altogether if the only
        # characters it has are digits.
        parts = flatten(s.strip() if (_exceptions_re.match(s)
                                      or (not s[0].isdigit() and s.endswith(_common_suffixes)))
                        else filter(bool, [t.lstrip(string.digits).rstrip(string.digits)
                                           for t in _numbers_re.split(s)])
                        for s in preliminary.split())
    return list(parts)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

from typing import List
from .javalang.tokenizer import JavaToken


class CodeCleaner:
    @staticmethod
    def clean_annotation(tokens:List[JavaToken]):
        cleaned_tokens = []
        size = len(tokens)
        index = 0
        flag = False
        left = 0
        right = 0
        while index < size:
            token = tokens[index]
            index += 1
            if len(token.value) == 0:
                continue
            if token.value == "@":
                flag = True
                index += 1
                continue
            if flag:
                if token.value == "(":
                    left += 1
                elif token.value == ")":
                    right += 1
                elif left == right:
                    flag = False
            if not flag:
                cleaned_tokens.append(token)
        return cleaned_tokens
            


#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import functools
import re

from .spiral import ronin

class Delimiter:

    # @staticmethod
    # @functools.lru_cache(maxsize=10000)
    # def split_camel(camel_case: str, to_lower=True):
    #     delimited_case = camel_case.replace('...', '-DOTDOTDOT-')
    #     delimited_case = delimited_case.split('.')[-1]
    #     delimited_case = delimited_case.replace('[]', ' []')
    #     delimited_case = delimited_case.replace('-DOTDOTDOT-', ' ...')
    #     delimited_case = re.sub(r'_', " ", delimited_case).strip()
    #     delimited_case = re.sub(
    #         r'([A-Za-z]| )([0-9]+)([A-Za-z]|$| )', r'\1 \2 \3', delimited_case)
    #     delimited_case = re.sub(r'([A-Z]+)([A-Z][a-z])', r'\1 \2', delimited_case)
    #     delimited_case = re.sub(r'([A-Z]+)', r' \1', delimited_case)
    #     delimited_case = re.sub(r'\s+', ' ', delimited_case)
    #     delimited_case = delimited_case.strip()

    #     if to_lower:
    #         delimited_case = delimited_case.lower()
    #     return delimited_case

    @staticmethod
    @functools.lru_cache(maxsize=10000)
    def split_camel(camel_case: str, to_lower=True):
        delimited_result: list = ronin.split(camel_case)
        # check ['...']
        if camel_case.endswith('...'): delimited_result.append('...')
        # check ['mult', 'i']
        size = len(delimited_result)
        for i in range(size):
            if i < size - 1 and delimited_result[i]=='mult' and delimited_result[i+1]=='i':
                pre = delimited_result[:i]
                pre.append('multi')
                pre.extend(delimited_result[i+2:])
                delimited_result = pre
                break
        delimited_case = ' '.join(delimited_result)
        
        if to_lower:
            delimited_case = delimited_case.lower()
        return delimited_case